@@ -28,6 +28,7 @@
 #endif
 // clang-format on
 
+#include <type_traits>
 #include "cpp11/altrep.hpp"
 
 namespace cpp11 {
@@ -48,8 +49,17 @@ template <typename T>
 inline T na();
 
 template <typename T>
-inline bool is_na(const T& value) {
+inline typename std::enable_if<!std::is_same<typename std::decay<T>::type, double>::value,
+                               bool>::type
+is_na(const T& value) {
   return value == na<T>();
 }
 
+template <typename T>
+inline typename std::enable_if<std::is_same<typename std::decay<T>::type, double>::value,
+                               bool>::type
+is_na(const T& value) {
+  return ISNA(value);
+}
+
 }  // namespace cpp11
