@@ -54,23 +54,35 @@ inline void set_option(SEXP name, SEXP value) {
   SETCAR(opt, value);
 }
 
-inline Rboolean& get_should_unwind_protect() {
+inline Rboolean* setup_should_unwind_protect() {
   SEXP should_unwind_protect_sym = Rf_install("cpp11_should_unwind_protect");
   SEXP should_unwind_protect_sexp = Rf_GetOption1(should_unwind_protect_sym);
+
   if (should_unwind_protect_sexp == R_NilValue) {
+    // Allocate and initialize once, then let R manage it.
+    // That makes this a shared global across all compilation units.
     should_unwind_protect_sexp = PROTECT(Rf_allocVector(LGLSXP, 1));
+    SET_LOGICAL_ELT(should_unwind_protect_sexp, 0, TRUE);
     detail::set_option(should_unwind_protect_sym, should_unwind_protect_sexp);
     UNPROTECT(1);
   }
 
-  Rboolean* should_unwind_protect =
-      reinterpret_cast<Rboolean*>(LOGICAL(should_unwind_protect_sexp));
-  should_unwind_protect[0] = TRUE;
-
-  return should_unwind_protect[0];
+  return reinterpret_cast<Rboolean*>(LOGICAL(should_unwind_protect_sexp));
 }
 
-static Rboolean& should_unwind_protect = get_should_unwind_protect();
+inline Rboolean* access_should_unwind_protect() {
+  // Setup is run once per compilation unit, but all compilation units
+  // share the same global option, so each compilation unit's static pointer
+  // will point to the same object.
+  static Rboolean* p_should_unwind_protect = setup_should_unwind_protect();
+  return p_should_unwind_protect;
+}
+
+inline Rboolean get_should_unwind_protect() { return *access_should_unwind_protect(); }
+
+inline void set_should_unwind_protect(Rboolean should_unwind_protect) {
+  *access_should_unwind_protect() = should_unwind_protect;
+}
 
 }  // namespace detail
 
@@ -82,11 +94,11 @@ static Rboolean& should_unwind_protect = get_should_unwind_protect();
 template <typename Fun, typename = typename std::enable_if<std::is_same<
                             decltype(std::declval<Fun&&>()()), SEXP>::value>::type>
 SEXP unwind_protect(Fun&& code) {
-  if (detail::should_unwind_protect == FALSE) {
+  if (detail::get_should_unwind_protect() == FALSE) {
     return std::forward<Fun>(code)();
   }
 
-  detail::should_unwind_protect = FALSE;
+  detail::set_should_unwind_protect(FALSE);
 
   static SEXP token = [] {
     SEXP res = R_MakeUnwindCont();
@@ -96,7 +108,7 @@ SEXP unwind_protect(Fun&& code) {
 
   std::jmp_buf jmpbuf;
   if (setjmp(jmpbuf)) {
-    detail::should_unwind_protect = TRUE;
+    detail::set_should_unwind_protect(TRUE);
     throw unwind_exception(token);
   }
 
@@ -121,7 +133,7 @@ SEXP unwind_protect(Fun&& code) {
   // unset it here before returning the value ourselves.
   SETCAR(token, R_NilValue);
 
-  detail::should_unwind_protect = TRUE;
+  detail::set_should_unwind_protect(TRUE);
 
   return res;
 }
@@ -316,16 +328,18 @@ static struct {
 
     static SEXP list_ = get_preserve_list();
 
-    // Add a new cell that points to the previous end.
-    SEXP cell = PROTECT(Rf_cons(list_, CDR(list_)));
+    // Get references to head, tail of the precious list.
+    SEXP head = list_;
+    SEXP tail = CDR(list_);
 
+    // Add a new cell that points to the current head + tail.
+    SEXP cell = PROTECT(Rf_cons(head, tail));
     SET_TAG(cell, obj);
 
-    SETCDR(list_, cell);
-
-    if (CDR(cell) != R_NilValue) {
-      SETCAR(CDR(cell), cell);
-    }
+    // Update the head + tail to point at the newly-created cell,
+    // effectively inserting that cell between the current head + tail.
+    SETCDR(head, cell);
+    SETCAR(tail, cell);
 
     UNPROTECT(2);
 
@@ -353,29 +367,25 @@ static struct {
 #endif
   }
 
-  void release(SEXP token) {
-    if (token == R_NilValue) {
+  void release(SEXP cell) {
+    if (cell == R_NilValue) {
       return;
     }
 
 #ifdef CPP11_USE_PRESERVE_OBJECT
-    R_ReleaseObject(token);
+    R_ReleaseObject(cell);
     return;
 #endif
 
-    SEXP before = CAR(token);
+    // Get a reference to the cells before and after the token.
+    SEXP lhs = CAR(cell);
+    SEXP rhs = CDR(cell);
 
-    SEXP after = CDR(token);
-
-    if (before == R_NilValue && after == R_NilValue) {
-      Rf_error("should never happen");
-    }
-
-    SETCDR(before, after);
-
-    if (after != R_NilValue) {
-      SETCAR(after, before);
-    }
+    // Remove the cell from the precious list -- effectively, we do this
+    // by updating the 'lhs' and 'rhs' references to point at each-other,
+    // effectively removing any references to the cell in the pairlist.
+    SETCDR(lhs, rhs);
+    SETCAR(rhs, lhs);
   }
 
  private:
@@ -415,18 +425,24 @@ static struct {
 
   static SEXP get_preserve_list() {
     static SEXP preserve_list = R_NilValue;
-
     if (TYPEOF(preserve_list) != LISTSXP) {
       preserve_list = get_preserve_xptr_addr();
       if (TYPEOF(preserve_list) != LISTSXP) {
-        preserve_list = Rf_cons(R_NilValue, R_NilValue);
+        preserve_list = Rf_cons(R_NilValue, Rf_cons(R_NilValue, R_NilValue));
         R_PreserveObject(preserve_list);
         set_preserve_xptr(preserve_list);
       }
+
+      // NOTE: Because older versions of cpp11 (<= 0.4.2) initialized the
+      // precious_list with a single cell, we might need to detect and update
+      // an existing empty precious list so that we have a second cell following.
+      if (CDR(preserve_list) == R_NilValue)
+        SETCDR(preserve_list, Rf_cons(R_NilValue, R_NilValue));
     }
 
     return preserve_list;
   }
-}  // namespace cpp11
-preserved;
+
+} preserved;
+
 }  // namespace cpp11
